import json
from concurrent.futures import ThreadPoolExecutor
from threading import Thread, RLock

from fga.exception import InvalidAPIKeyError
from fga.http_client import HttpClient
from fga.processor import ResponseProcessor
from fga.constants import DEFAULT_HEADER


class Workers:

    def __init__(self):
        self.threads_pool = ThreadPoolExecutor(max_workers=16)
        self.is_active = True
        self.consumer_lock = RLock()
        self.is_started = False
        self.configuration = None
        self.storage = None
        self.response_processor = ResponseProcessor()

    def setup(self, configuration, storage):
        self.configuration = configuration
        self.storage = storage
        self.response_processor.setup(configuration, storage)

    def start(self):
        with self.consumer_lock:
            if not self.is_started:
                self.is_started = True
                consumer = Thread(target=self.buffer_consumer)
                consumer.start()

    def stop(self):
        self.flush()
        self.is_active = False
        self.is_started = True
        self.threads_pool.shutdown()

    def flush(self):
        events = self.storage.pull_all()
        if events:
            return self.threads_pool.submit(self.send, events)

    def send(self, events):
        url = self.configuration.server_url
        print('url', url)
        payload = self.get_payload(events)
        print('payload', payload)
        DEFAULT_HEADER["token"] = self.configuration.api_key
        DEFAULT_HEADER["project-id"] = self.configuration.project_id
        res = HttpClient.post(url, payload, header=DEFAULT_HEADER)
        print('Send Res', res.body)

        try:
            self.response_processor.process_response(res, events)
        except InvalidAPIKeyError as  e:
            print('InvalidAPIKeyError', e)
            self.configuration.logger.error("Invalid API Key")

    def get_payload(self, events) -> bytes:
        payload_body = []
        for event in events:
            event_body = event.get_event_body()
            if event_body:
                payload_body.append(event_body)
        # if self.configuration.options:
        #     payload_body["options"] = self.configuration.options
        return json.dumps(payload_body, sort_keys=True).encode('utf8')

    def buffer_consumer(self):
        try:
            if self.is_active:
                with self.storage.lock:
                    self.storage.lock.wait(self.configuration.flush_interval_millis / 1000)
                    while True:
                        if not self.storage.total_events:
                            break
                        events = self.storage.pull(self.configuration.flush_queue_size)
                        if events:
                            self.threads_pool.submit(self.send, events)
                        else:
                            wait_time = self.storage.wait_time / 1000
                            if wait_time > 0:
                                self.storage.lock.wait(wait_time)
        except Exception:
            self.configuration.logger.exception("Consumer thread error")
        finally:
            with self.consumer_lock:
                self.is_started = False



"""Amplitude client module that provide client class with public interface and configuration attributes

Classes:
    Amplitude: the Amplitude client class
"""

from typing import Optional, Union, List

from fga import constants
from fga.config import Config
from fga.event import Revenue, BaseEvent, Identify, IdentifyEvent, GroupIdentifyEvent, EventOptions, User, Event
from fga.plugin import FGADestinationPlugin, ContextPlugin, Plugin, EnrichmentPlugin, UserEnrichmentPlugin
from fga.timeline import Timeline
from fga.exception import InvalidEventError
import atexit
import threading


class FGA:
    """Amplitude client used to store configurations and track events

    Args:
        api_key (str): The api key of amplitude project. Must be set properly before tracking events.
        configuration (amplitude.config.Config, optional): The configuration of client instance. A new instance
            with default config value will be used by default.

    Attributes:
        configuration (amplitude.config.Config): the configuration of client instance

    Methods:
        track(event): Process and send event
        identify(identify_obj, event_properties, event_options): Send an identify event to update user properties
        group_identify(group_type, group_name, identify_obj, event_options, event_properties, user_properties): Send
            a group identify event to update group properties
        revenue(revenue_obj, event_options): Send a revenue event with revenue info in event_properties
        flush(): Flush all event waiting to be sent in the buffer
        add(plugin): Add the plugin object to client instance.
        remove(plugin): Remove the plugin object from client instance
        shutdown(): Shutdown the client instance
    """

    def __init__(self, token: str, project_id: int, configuration: Optional[Config] = None):
        """The constructor for the Amplitude class

        Args:
            api_key (str): The api key of amplitude project. Must be set properly before tracking events.
            configuration (amplitude.config.Config, optional): The configuration of client instance. A new instance
                with default config value will be used by default.
        """
        self.configuration: Config = configuration or Config()
        self.configuration.api_key = token
        self.configuration.project_id = project_id
        self.__timeline = Timeline()
        self.__timeline.setup(self)
        self._register_on_exit()
        self.add(FGADestinationPlugin())
        self.add(UserEnrichmentPlugin())
        self.add(EnrichmentPlugin())
        self.add(ContextPlugin())

    def track(self, event: BaseEvent):
        """Process and send the given event object.

        Args:
            event (amplitude.event.BaseEvent): The event that we want to track
        """
        if isinstance(event, Event):
            self.configuration.server_url = constants.SERVER_URL[constants.EVENT]
        elif isinstance(event, User):
            self.configuration.server_url = constants.SERVER_URL[constants.USER]
        else:
            raise InvalidEventError("Invalid event.")
        self.__timeline.process(event)

    def identify(self, identify_obj: Identify, event_options: EventOptions, event_properties: Optional[dict] = None):
        """Send an identify event to update user properties

        Args:
            identify_obj (amplitude.event.Identify): Identify object contain operations of updating user properties
            event_options (amplitude.event.EventOptions): Provide additional information to identify event.
            event_properties (dict, optional): A dictionary of event properties.
        """
        if not identify_obj.is_valid():
            self.configuration.logger.error("Empty identify properties")
        else:
            event = IdentifyEvent(event_properties=event_properties,
                                  user_properties=identify_obj.user_properties)
            event.load_event_options(event_options)
            self.track(event)

    def user_set(self, user_id: str, user_properties: Optional[dict] = None):
        if user_properties is None:
            self.configuration.logger.error("Empty identify properties")
        else:
            event = User(user_id=user_id, user_properties=user_properties)
            self.track(event)

    def user_set_once(self, user_id: str, key: str, value):
        user_properties = {}
        user_properties[key] = value
        event = User(user_id=user_id, user_properties=user_properties, event_properties={'user_operation': constants.IDENTITY_OP_SET_ONCE, 'set_once_key': key })
        self.track(event)

    def group_identify(self, group_type: str, group_name: str, identify_obj: Identify,
                       event_options: Optional[EventOptions] = None,
                       event_properties: Optional[dict] = None,
                       user_properties: Optional[dict] = None):
        """Send a group identify event to update group properties

        Args:
            group_type (str): The group type e.g. "sport"
            group_name (str): The group name e.g. "soccer"
            identify_obj (amplitude.event.Identify): Identify object contain operations of updating group properties
            event_options (amplitude.event.EventOptions, optional): Provide additional information to
                group identify event like user_id.
            event_properties (dict, optional): A dictionary of event properties. Defaults to None.
            user_properties (dict, optional): A dictionary of user properties. Defaults to None.
        """
        if not identify_obj.is_valid():
            self.configuration.logger.error("Empty group identify properties")
        else:
            event = GroupIdentifyEvent(event_properties=event_properties,
                                       user_properties=user_properties,
                                       groups={group_type: group_name},
                                       group_properties=identify_obj.user_properties)
            event.load_event_options(event_options)
            self.track(event)

    def revenue(self, revenue_obj: Revenue, event_options: EventOptions):
        """Send a revenue event with revenue info in event_properties

        Args:
            revenue_obj (amplitude.event.Revenue): A revenue object that contains information like price,
                quantity, receipt,  revenue_type
            event_options (amplitude.event.EventOptions): Provide additional information to revenue event
                like user_id.
        """
        if not revenue_obj.is_valid():
            self.configuration.logger.error("Invalid price for revenue event")
        else:
            event = revenue_obj.to_revenue_event()
            event.load_event_options(event_options)
            self.track(event)

    def set_group(self, group_type: str, group_name: Union[str, List[str]], event_options: EventOptions):
        """Sending an identify event to put a user in group(s) by setting group type and group name as
            user property for a user.

        Args:
            group_type (str): The group type e.g. "sport"
            group_name (str): The group name e.g. "soccer"
            event_options (amplitude.event.EventOptions): Provide additional information for event
                like user_id.
        """
        identify = Identify().set(group_type, group_name)
        event = IdentifyEvent(groups={group_type: group_name}, user_properties=identify.user_properties)
        event.load_event_options(event_options)
        self.track(event)

    def flush(self):
        """Flush all event waiting to be sent in the buffer

        Returns:
            A list of Future objects for all destination plugins
        """
        return self.__timeline.flush()

    def add(self, plugin: Plugin):
        """Add the plugin object to client instance. Events tracked by this client instance will be
            processed by instance's plugins.

        Args:
            plugin (amplitude.plugin.Plugin): the plugin object to be added to the client instance

        Returns:
            Amplitude: the client instance itself
        """
        self.__timeline.add(plugin)
        plugin.setup(self)
        return self

    def remove(self, plugin: Plugin):
        """Remove the plugin object from client instance

        Args:
            plugin (amplitude.plugin.Plugin): the plugin object to be removed from the client instance

        Returns:
            Amplitude: the client instance itself
        """
        self.__timeline.remove(plugin)
        return self

    def shutdown(self):
        """Shutdown the client instance, not accepting new events, flush all events in buffer"""
        self.configuration.opt_out = True
        self.__timeline.shutdown()

    def _register_on_exit(self):
        """Internal method to clean up the client instance on main thread exit"""
        if hasattr(threading, "_register_atexit"):
            try:
                threading._register_atexit(self.shutdown)
            except Exception as e:
                self.configuration.logger.warning("register for exit fail")
        else:
            atexit.register(self.shutdown)



"""Amplutide plugin module. Provide base class to implement customized plugin

Classes:
    Plugin: Base class of all plugins.
    EventPlugin: Base class to implement plugins that modify and enrich events.
    DestinationPlugin: Base class to implement plugins that send events to customized destinations.
    AmplitudeDestinationPlugin: Default Amplitude Destination plugin that send events to Amplitude.
    ContextPlugin: A default plugin that add library info to event. Also set event default timestamp and insert_id
        if not set elsewhere.

Methods:
    verify_event(event): Perform basic validation before AmplitudeDestinationPlugin send the event to storage.
"""
import abc
import copy
import json
import time

import requests
import pydash
import uuid
from typing import Optional

from fga.event import BaseEvent, GroupIdentifyEvent, IdentifyEvent, RevenueEvent, User
from fga import constants
from fga.timeline import Timeline
from fga.exception import InvalidEventError
from fga import utils
from fga.worker import Workers
from fga.constants import DEFAULT_HEADER, UserPropertiesOperation


class Plugin(abc.ABC):
    """The abstract base class of plugins

    Args:
        plugin_type (constants.PluginType): The plugin type.
    """

    def __init__(self, plugin_type: constants.PluginType):
        """The constructor of Plugin class"""
        self.plugin_type: constants.PluginType = plugin_type

    def setup(self, client):
        """Setup plugins with client instance parameter"""
        pass

    @abc.abstractmethod
    def execute(self, event: BaseEvent):
        """Process event with plugin instance"""
        pass


class EventPlugin(Plugin):
    """Plugins that modify and enrich events. Used as base class of event plugins.

    Args:
        plugin_type (constants.PluginType): The plugin type.

    Methods:
        setup(client): Setup plugin using Amplitude client instance.
        execute(event): Method to override to process event. Return modified event or None. Return None will stop the
            Amplitude client sending the event and callback will not be triggered.
        track(event): Can be override to process BaseEvent if execute method not overrided.
        revenue(event): Can be override to process RevenueEvent if execute method not overrided.
        identify(event): Can be override to process IdentifyEvent if execute method not overrided.
        group_identify(event): Can be override to process GroupIdentifyEvent if execute method not overrided.
    """

    def __init__(self, plugin_type: constants.PluginType):
        super().__init__(plugin_type)

    def execute(self, event: BaseEvent) -> Optional[BaseEvent]:
        if isinstance(event, GroupIdentifyEvent):
            return self.group_identify(event)
        if isinstance(event, IdentifyEvent):
            return self.identify(event)
        if isinstance(event, RevenueEvent):
            return self.revenue(event)
        return self.track(event)

    def group_identify(self, event: GroupIdentifyEvent) -> Optional[GroupIdentifyEvent]:
        return event

    def identify(self, event: IdentifyEvent) -> Optional[IdentifyEvent]:
        return event

    def revenue(self, event: RevenueEvent) -> Optional[RevenueEvent]:
        return event

    def track(self, event: BaseEvent) -> Optional[BaseEvent]:
        return event


class UserEnrichmentPlugin(EventPlugin):

    def __init__(self):
        super().__init__(constants.PluginType.ENRICHMENT)
        self.configuration = None
        self.url = constants.SERVER_URL[constants.USER]

    def setup(self, client):
        super().setup(client)
        self.configuration = client.configuration

    def execute(self, event: BaseEvent) -> None:
        if not isinstance(event, User):
            return event
        default_header = copy.deepcopy(DEFAULT_HEADER)
        default_header["token"] = self.configuration.api_key
        default_header["project-id"] = str(self.configuration.project_id)
        get_use_url = f'{self.url}/{event.user_id}'
        r = requests.get(get_use_url, headers=default_header)
        result = r.json()
        print(result)
        if len(pydash.get(result, 'data', [])) == 0:
            return event
        current_user = pydash.get(result, 'data.0')
        if pydash.get(event, 'event_properties.user_operation') == constants.IDENTITY_OP_SET_ONCE \
                and pydash.get(current_user, pydash.get(event, 'event_properties.set_once_key')) is not None:
            return None
        user_properties = {}
        for key, value in current_user.items():
            user_properties[key] = value if pydash.get(event, f'user_properties.{key}') is None else event.user_properties[key]
        event['user_properties'] = user_properties
        return event


class EnrichmentPlugin(EventPlugin):

    def __init__(self):
        super().__init__(constants.PluginType.ENRICHMENT)
        self.configuration = None

    def setup(self, client):
        self.configuration = client.configuration

    def execute(self, event: BaseEvent):
        event["project_id"] = self.configuration.project_id
        return event


class DestinationPlugin(EventPlugin):
    """Plugins that send events to a destination like Amplitude.

    Methods:
        setup(client): Setup plugin using Amplitude client instance.
        execute(event): Method to override to send out events.
        add(plugin): Add additional processing ability by plugins to modify events before sending out.
        remove(plugin): Remove a plugin instance from destination plugin.
        shutdown(): Method to override to handle closure of client like flushing events,
            closing threads and connections. Triggered by client.shutdown()
    """

    def __init__(self):
        super().__init__(constants.PluginType.DESTINATION)
        self.timeline = Timeline()

    def setup(self, client):
        self.timeline.setup(client)

    def add(self, plugin):
        self.timeline.add(plugin)
        return self

    def remove(self, plugin):
        self.timeline.remove(plugin)
        return self

    def execute(self, event: BaseEvent) -> None:
        event = self.timeline.process(event)
        super().execute(event)

    def shutdown(self):
        self.timeline.shutdown()


class FGADestinationPlugin(DestinationPlugin):
    """The Amplitude destination plugin. Added to client by default. Send events to Amplitude.

    Methods:
        setup(client): Setup plugin instance and storage and workers instance of the destination plugin.
        execute(event): Process event with plugins added to the destination plugin. Then pushed the event to storage
            waiting to be sent.
        flush(): Flush all event in storage instance.
        shutdown(): Shutdown plugins and works of the destination plugin.
    """

    def __init__(self):
        """The constructor of AmplitudeDestinationPlugin class"""
        super().__init__()
        self.workers = Workers()
        self.storage = None
        self.configuration = None

    def setup(self, client):
        """Setup plugin instance and storage and workers instance of the destination plugin.

        Args:
            client: The Amplitude client that holds the destination plugin.
        """
        super().setup(client)
        self.configuration = client.configuration
        self.storage = client.configuration.get_storage()
        self.workers.setup(client.configuration, self.storage)
        self.storage.setup(client.configuration, self.workers)

    def execute(self, event: BaseEvent) -> None:
        """Process event with plugins added to the destination plugin. Then pushed the event to storage
            waiting to be sent.

        Args:
            event (BaseEvent): The event to be sent.
        """
        event = self.timeline.process(event)
        print('plugin event', event)
        if not verify_event(event):
            raise InvalidEventError("Invalid event.")
        self.storage.push(event)

    def flush(self):
        """Flush all event in storage instance."""
        return self.workers.flush()

    def shutdown(self):
        """Shutdown plugins and works of the destination plugin."""
        self.timeline.shutdown()
        self.workers.stop()


class ContextPlugin(Plugin):
    """Amplitude Context plugin. Default added to client. Add library info to event.
        Also set event default timestamp and insert_id if not set elsewhere.

    Methods:
        apply_context_data(event): Add SDK name and version to event.library.
        execute(event):
            - Set event default timestamp and insert_id if not set elsewhere.
            - Add SDK name and version to event.library.
            - Mount plan, ingestion_metadata if not yet.
    """

    def __init__(self):
        """The constructor of ContextPlugin class"""
        super().__init__(constants.PluginType.BEFORE)
        self.context_string = f"{constants.SDK_LIBRARY}/{constants.SDK_VERSION}"
        self.configuration = None

    def setup(self, client):
        self.configuration = client.configuration

    def apply_context_data(self, event: BaseEvent):
        """Add SDK name and version to event.library.

        Args:
            event (BaseEvent): The event to be processed.
        """
        event.library = self.context_string

    def execute(self, event: BaseEvent) -> BaseEvent:
        """
        - Set event default timestamp and insert_id if not set elsewhere.
        - Add SDK name and version to event.library.
        - Mount plan, ingestion_metadata if not yet.

        Args:
            event (BaseEvent): The event to be processed.
        """
        if not event.time:
            event["time"] = int(time.time())
        if not event.insert_id:
            event["insert_id"] = str(uuid.uuid4())
        if not event.event_id:
            event["event_id"] = event["insert_id"]
        if self.configuration.plan and (not event.plan):
            event["plan"] = self.configuration.plan
        if self.configuration.ingestion_metadata and (not event.ingestion_metadata):
            event["ingestion_metadata"] = self.configuration.ingestion_metadata
        self.apply_context_data(event)
        return event


def verify_event(event):
    """Perform basic validation before AmplitudeDestinationPlugin send the event to storage.

    Args:
        event (BaseEvent): the event to be verified.

    Returns:
        True is event is valid, False otherwise.
    """
    if isinstance(event, GroupIdentifyEvent):
        return True
    if (not event["event_type"]) or \
            (not event["user_id"] and not event["device_id"]):
        return False
    return True



import logging
import time

from fga import constants

logger = logging.getLogger(constants.LOGGER_NAME)


def current_milliseconds() -> int:
    """Return system current time in milliseconds

    Returns:
         current time in milliseconds in integer
    """
    return int(time.time() * 1000)


def truncate(obj):
    """Truncate an object: Cut string values that exceed the maximum length limit. Return empty dict if number of
        items exceed maximum properties key limit. Truncate operation applies to dict and str type, values of
        dict and elements of list.

    Args:
         obj: The object to be truncated.

    Returns:
        The truncated object
    """
    if isinstance(obj, dict):
        if len(obj) > constants.MAX_PROPERTY_KEYS:
            logger.error(f"Too many properties. {constants.MAX_PROPERTY_KEYS} maximum.")
            return {}
        for key, value in obj.items():
            obj[key] = truncate(value)
    elif isinstance(obj, list):
        for i, element in enumerate(obj):
            obj[i] = truncate(element)
    elif isinstance(obj, str):
        obj = obj[:constants.MAX_STRING_LENGTH]
    return obj
